// dynamic.cpp
// Revision 21-dec-2003

#include "dynamic.h"
#include "error.h"

#if (defined __unix__ || defined __linux__ || defined __NetBSD__) &&  \
	! defined __CYGWIN__
// Kylix defines only __linux__

#ifdef __hpux__

#include <dl.h>

#if 0
typedef shl_t DynamicHandle;

inline DynamicHandle dynamicload (const std::string & str)
	{ return shl_load (str.c_str (), BIND_DEFERRED, 0); }

inline void * dynamicaddr (DynamicHandle & handle, const std::string & str)
{
	void * value;
	if (shl_findsym (& handle, str.c_str (), TYPE_UNDEFINED, & value) == 0)
		return value;
	else return NULL;
}

inline void dynamicclose (DynamicHandle handle)
	{ shl_unload (handle); }
#else

class DynamicHandle::Internal {
	shl_t handle;
public:
	Internal (const std::string & name)
	{
		handle= shl_load (name.c_str (), BIND_DEFERRED, 0);
		if (handle == NULL)
			throw ErrNoDynamicLibrary;
	}
	~Internal ()
	{
		// HP-UX seems to unload the library even if it was
		// opened several times, then we don't un unload it.
		//shl_unload (handle);
	}
	DynamicUsrFunc addr (const std::string & str)
	{
		void * value;
		if (shl_findsym (& handle, str.c_str (), TYPE_UNDEFINED,
				& value) == 0)
		{
			//return reinterpret_cast <DynamicUsrFunc> (value);
			return (DynamicUsrFunc) (value);
		}
		else
			throw ErrNoDynamicSymbol;
	}
};

#endif

#else

#include <dlfcn.h>

#if 0
typedef void * DynamicHandle;

inline DynamicHandle dynamicload (const std::string & str)
        { return dlopen (str.c_str (), RTLD_LAZY); }

inline void * dynamicaddr (DynamicHandle handle, const std::string & str)
        { return dlsym (handle, str.c_str () ); }

inline void dynamicclose (DynamicHandle handle)
        { dlclose (handle); }
#else

class DynamicHandle::Internal {
	void * handle;
public:
	Internal (const std::string & name)
	{
		handle= dlopen (name.c_str (), RTLD_LAZY);
		if (handle == NULL)
			throw ErrNoDynamicLibrary;
	}
	~Internal ()
	{
		dlclose (handle);
	}
	DynamicUsrFunc addr (const std::string & str)
	{
		void * value;
		if ( (value= dlsym (handle, str.c_str () ) ) != NULL)
		{
			//return reinterpret_cast <DynamicUsrFunc> (value);
			return (DynamicUsrFunc) value;
		}
		else
			throw ErrNoDynamicSymbol;
	}
};

#endif

#endif

#elif defined __WIN32__ || defined __CYGWIN__

#include <windows.h>
#undef min
#undef max

#if 0
typedef HMODULE DynamicHandle;

inline DynamicHandle dynamicload (const std::string & str)
        { return LoadLibrary (str.c_str () ); }

inline void * dynamicaddr (DynamicHandle handle, const std::string & str)
{
	return (void *) GetProcAddress (handle, str.c_str () );
}

inline void dynamicclose (DynamicHandle handle)
        { FreeLibrary (handle); }
#else

class DynamicHandle::Internal {
	HMODULE handle;
public:
	Internal (const std::string & name)
	{
		handle= LoadLibrary (name.c_str () );
		if (handle == NULL)
			throw ErrNoDynamicLibrary;
	}
	~Internal ()
	{
		FreeLibrary (handle);
	}
	DynamicUsrFunc addr (const std::string & str)
	{
		void * value;
		if ( (value= (void *)
			GetProcAddress (handle, str.c_str () ) ) != NULL)
		{
			//return reinterpret_cast <DynamicUsrFunc> (value);
			return (DynamicUsrFunc) value;
		}
		else
			throw ErrNoDynamicSymbol;
	}
};

#endif

#else

//#error Unknown operating system
#warning Dynamic link unsupported

class DynamicHandle::Internal {
public:
	Internal (const std::string & name)
	{
		throw "Dynamic link unsupported";
	}
	~Internal ()
	{
	}
	DynamicUsrFunc addr (const std::string & str)
	{
		throw "Dynamic link unsupported";
	}
};


#endif


DynamicHandle::DynamicHandle () :
	pin (0)
{
}

DynamicHandle::DynamicHandle (const std::string & name) :
	pin (new Internal (name) )
{
}

void DynamicHandle::assign (const std::string & name)
{
	// Create the new before destructing the old, thus leaving
	// intouched in case of exception.
	Internal * npin= new Internal (name);
	delete pin;
	pin= npin;
}

DynamicHandle::~DynamicHandle ()
{
	delete pin;
}

DynamicUsrFunc DynamicHandle::addr (const std::string & str)
{
	return pin->addr (str);
}

// Fin de dymanic.cpp
